(* BASIC DEFINITIONS *)

(* Reserved keys *)
let keys = [
  "gra"; "-graphical";
  "par"; "-parallel";
  "seq"; "-sequential";
  "pro"; "-process";
  "thr"; "-thread";
  "gra"; "-graphical";
  "v"; "-version";
  "h"; "help"; "-help";
];;

type exe =
   | Byt | Bin | Top;;
type semantic =
   | Seq | Par_pro | Par_thr | Gra ;;

(* Semantics evaluation association list =
   sem * (sem_name * sem_path_list * sem_ext_lib) *)
let sem_assoc = [
    (Seq, ("sequential",
           [Filename.concat "+camlp3l" "seq"],
           []));
    (Gra, ("graphical",
           [Filename.concat "+camlp3l" "gra"],
           ["graphics"]));
    (Par_pro, ("parallel<process>",
               [Filename.concat "+camlp3l" (Filename.concat "par" "pro");
                Filename.concat "+camlp3l" "par"],
               ["unix"]));
    (Par_thr, ("parallel<thread>",
               [Filename.concat "+camlp3l" (Filename.concat "par" "thr");
                Filename.concat "+camlp3l" "par";
                "+threads"],
               ["unix"; "threads"]));
];;

(* Executable kind association list =
   exe * (exe_name * exe_cmd * exe_sfx_obj * exe_sfx_lib) *)
let exe_assoc = [
    (Byt, ("bytecode compiler", "ocamlc", "cmo", "cma"));
    (Bin, ("nativecode compiler", "ocamlopt", "cmx", "cmxa"));
    (Top, ("toplevel (bytecode)", "ocaml", "cmo", "cma"))
];;

(* Configuration *)
type config = {
  mutable exe : exe;
  mutable sem : semantic;
  mutable args : string list;
  mutable path : string;
};;

(* default configuration *)
let default_configuration = {
  exe = Byt;
  sem = Seq;
  args = [];
  path = Filename.concat "+camlp3l" "seq";
};;

(* Setting the semantics evaluation regime choice. *)
let set_seq () = default_configuration.sem <- Seq;;
let set_par_pro () = default_configuration.sem <- Par_pro;;
let set_par_thr () = default_configuration.sem <- Par_thr;;
let set_gra () = default_configuration.sem <- Gra;;

(* Bulding the command line options for the semantics evaluation regime. *)
let make_seq_option opt_name =
  (opt_name, Arg.Unit set_seq, " Use sequential execution");;
let make_par_pro_option opt_name =
  (opt_name,
   Arg.Unit set_par_pro,
   " Use parallel execution with process model");;
let make_par_thr_option opt_name =
  (opt_name,
   Arg.Unit set_par_thr,
   " Use parallel execution with thread model");;
let make_gra_option opt_name =
  (opt_name, Arg.Unit set_gra, " Use graphical execution");;

(* Instantiate configuration *)
let set_args () =
  let argv_m = Array.of_list
    (List.fold_left
       (fun l s ->
          if String.get s 0 = '-' then
            let key = String.sub s 1 (String.length s - 1) in
            if not (List.mem key keys) then l @ ["-ext"; key] 
            else l @ [s]
          else l @ [s])
       [] (Array.to_list Sys.argv)
    ) in

  let speclist = [
      make_gra_option "-gra"; make_gra_option "--graphical";
      make_par_pro_option "-par"; make_par_pro_option "--parallel";
      make_par_pro_option "-pro"; make_par_pro_option "--process";
      make_par_thr_option "-thr"; make_par_thr_option "--thread";
      make_seq_option "-seq"; make_seq_option "--sequential";

      ("-ext",
       Arg.String
         (fun s ->
            default_configuration.args <-
              default_configuration.args @ ["-" ^ s]),
       " Internal use only")
  ] @ Command_options.all () in
  try
    Arg.parse_argv argv_m speclist
      (fun s -> default_configuration.args <- default_configuration.args @ [s])
      "CamlP3l can also accept most compiler options of OCaml."

  with
  | Arg.Bad msg -> Printf.eprintf "%s" msg; exit 2;
  | Arg.Help msg -> Printf.printf "%s" msg; exit 0;;

(* Add suffix string [suff] to filename [fname]. *)
let add_suffix suff fname = Printf.sprintf "%s.%s" fname suff;;

(* A command is constructed by the following parts in the following
   sequence:
   Cmd(exe) + ModOpt(model) + ModPath(model) + ModLib(model+exe) +
   SemLib(sem+exe) + SemObj(sem+exe) + UsrCmd *)
let get_cmd () =
  let exe_name, exe_cmd, exe_sfx_obj, exe_sfx_lib =
    List.assoc default_configuration.exe exe_assoc in
  let sem_name, sem_path_list, sem_ext_lib =
    List.assoc default_configuration.sem sem_assoc in
  let rflatten l =
    let rec aux accu = function
      | [] -> accu
      | [] :: t -> aux accu t
      | (hh :: ht) :: t ->
        aux (if List.mem hh accu then accu else accu @ [hh]) (ht :: t) in
    aux [] l in
  (* We automatically add the root directory where camlp3l is installed. *)
  let all_inc_path =
    List.flatten
      (List.map
         (fun s -> ["-I"; s])
         (Installation_data.camlp3l_library_path :: sem_path_list)) in
  let all_ext_lib =
    List.map (add_suffix exe_sfx_lib)
      (rflatten [sem_ext_lib; ["p3l"]]) in
  let argv_lib = all_inc_path @ all_ext_lib in
  let argv_usr =
    let argv_usr_dup_removed =
      List.filter
        (fun x -> not (List.mem x all_ext_lib))
        default_configuration.args in
    List.fold_left
      (fun l x ->
         if Filename.check_suffix x "3l"  then l @ ["-impl"; x] else
         if Filename.check_suffix x "3li" then l @ ["-intf"; x] else
         l @ [x]
      ) [] argv_usr_dup_removed in
  let argv_dst =
    let rec guess_exe_name = function
      | [] -> None
      | h :: _ when
          List.exists (Filename.check_suffix h) ["cmo"; "cmx"] ->
          Some (Filename.chop_extension h)
      | h1 :: h2 :: _ when h2 = "-impl" ->
          Some (Filename.chop_extension h1)
      | _ :: t -> guess_exe_name t in
    let exe_sfx_dst = "opt" in
    let sem_sfx_dst = String.sub sem_name 0 3 in
    if List.exists (fun x -> x = "-c" || x = "-o") argv_usr then []
    else match default_configuration.exe with
      | Top -> []
      | Byt ->
          (match guess_exe_name (List.rev argv_usr) with
          | None -> []
          | Some x -> ["-o"; add_suffix sem_sfx_dst x])
      | Bin ->
          (match guess_exe_name (List.rev argv_usr) with
           | None -> []
           | Some x ->
               ["-o"; String.concat "." [x; sem_sfx_dst; exe_sfx_dst]])
  in
  let desc =
    Printf.sprintf "Launch %s with %s semantics." exe_name sem_name in
  (exe_cmd, argv_lib @ argv_usr @ argv_dst, desc)
;;

let make_open_cmd mods =
  List.map (fun x -> Printf.sprintf "open %s\\;\\;" x) mods;;

let add_version_options bn =
  let si =
    {Version.software_info with Versionning.software_banner_name = bn} in
  Versionning.add_version_options si
;;

let make_main exe bn =
  add_version_options bn;
  let command, arguments, description =
    set_args ();
    default_configuration.exe <- exe;
    get_cmd () in

  let open_cmd = make_open_cmd ["P3l"] in

  let preproc_cmd =
    ["-pp "; "\"echo -n " ^ String.concat "" open_cmd ^ " | cat - \""] in

  let cmd = String.concat " " (command :: preproc_cmd @ arguments) in

  let main () =
    print_endline description;
    match Sys.command cmd with
    | 0 -> exit 0
    | n ->
      print_endline (
        Printf.sprintf
          "Command %s\nfailed with status %d, something went wrong."
          cmd n); 
      exit 2 in
  main;;

(** Other useful tools. *)

let get_path_list path =
  let rec split s =
    try
      let pos = String.index s ':' in
      String.sub s 0 pos ::
      split (String.sub s (pos + 1) ((String.length s) - pos - 1)) with
    | Not_found -> [s] in
  try split (Sys.getenv path) with
  | Not_found -> [];;
(** Get the user's system environment variable env. *)

let make_top_main exe bn =
  add_version_options bn;
  let command, arguments, description =
    set_args ();
    default_configuration.exe <- exe;
    get_cmd () in

  let init_file_name = Filename.temp_file "camlp3l" ".ml" in

  let build_init_file () =
    let oc = open_out init_file_name in
    List.iter (fun m ->
      output_string oc (Printf.sprintf "open %s;;\n" m))
      ["P3l"];
    output_string oc (Printf.sprintf "Sys.remove %S;;\n" init_file_name);
    close_out oc in

  let cmd =
    String.concat " " ([command; "-init"; init_file_name] @ arguments) in

  let main () =
    build_init_file ();
    print_endline description;
    match Sys.command cmd with
    | 0 -> exit 0
    | n ->
      print_endline (
        Printf.sprintf
          "Command %s\nfailed with status %d, something went wrong."
          cmd n); 
      exit 2 in

   main;;
